@@ -37,6 +37,7 @@
 #include "msc_lua.h"
 #endif
 
+#include "msc_status_engine.h"
 
 /* ModSecurity structure */
 
@@ -60,9 +61,17 @@ unsigned long int DSOLOCAL msc_pcre_match_limit = 0;
 
 unsigned long int DSOLOCAL msc_pcre_match_limit_recursion = 0;
 
+int DSOLOCAL status_engine_state = STATUS_ENGINE_DISABLED;
+
+int DSOLOCAL conn_limits_filter_state = MODSEC_DISABLED;
+
 unsigned long int DSOLOCAL conn_read_state_limit = 0;
+TreeRoot DSOLOCAL *conn_read_state_whitelist = 0;
+TreeRoot DSOLOCAL *conn_read_state_suspicious_list = 0;
 
 unsigned long int DSOLOCAL conn_write_state_limit = 0;
+TreeRoot DSOLOCAL *conn_write_state_whitelist = 0;
+TreeRoot DSOLOCAL *conn_write_state_suspicious_list = 0;
 
 #if defined(WIN32) || defined(VERSION_NGINX)
 int (*modsecDropAction)(request_rec *r) = NULL;
@@ -195,6 +204,7 @@ int perform_interception(modsec_rec *msr) {
             break;
 
         case ACTION_PROXY :
+#if !(defined(VERSION_IIS)) && !(defined(VERSION_NGINX)) && !(defined(VERSION_STANDALONE))
             if (msr->phase < 3) {
                 if (ap_find_linked_module("mod_proxy.c") == NULL) {
                     log_level = 1;
@@ -219,6 +229,15 @@ int perform_interception(modsec_rec *msr) {
                     "(Configuration Error: Proxy action requested but it does not work in output phases).",
                     phase_text);
             }
+#else
+            log_level = 1;
+            status = HTTP_INTERNAL_SERVER_ERROR;
+            message = apr_psprintf(msr->mp, "Access denied with code 500%s "
+                "(Configuration Error: Proxy action to %s requested but "
+                "proxy is only available in Apache version).",
+                phase_text,
+                log_escape_nq(msr->mp, actionset->intercept_uri));
+#endif
             break;
 
         case ACTION_DROP :
@@ -509,6 +528,8 @@ static modsec_rec *create_tx_context(request_rec *r) {
     msr->request_headers = apr_table_copy(msr->mp, r->headers_in);
     msr->hostname = ap_get_server_name(r);
 
+    msr->msc_full_request_buffer = NULL;
+    msr->msc_full_request_length = 0;
     msr->msc_rule_mptmp = NULL;
 
     /* Invoke the engine to continue with initialisation */
@@ -537,6 +558,11 @@ static modsec_rec *create_tx_context(request_rec *r) {
 static apr_status_t change_server_signature(server_rec *s) {
     char *server_version = NULL;
 
+    /* This is a very particular way to handle the server banner. It is Apache
+     * only. Stanalone and descendants should address that in its specifics
+     * implementations, e.g. Nginx module.
+     */
+#if !(defined(VERSION_IIS)) && !(defined(VERSION_NGINX)) && !(defined(VERSION_STANDALONE))
     if (new_server_signature == NULL) return 0;
 
     server_version = (char *)apache_get_server_version();
@@ -568,7 +594,7 @@ static apr_status_t change_server_signature(server_rec *s) {
     else {
         ap_log_error(APLOG_MARK, APLOG_DEBUG | APLOG_NOERRNO, 0, s, "SecServerSignature: Changed server signature to \"%s\".", server_version);
     }
-
+#endif
     return 1;
 }
 
@@ -705,6 +731,17 @@ static int hook_post_config(apr_pool_t *mp, apr_pool_t *mp_log, apr_pool_t *mp_t
             ap_log_error(APLOG_MARK, APLOG_NOTICE | APLOG_NOERRNO, 0, s,
                     "Original server signature: %s", real_server_signature);
         }
+
+#ifndef WIN32
+        if (status_engine_state != STATUS_ENGINE_DISABLED) {
+            msc_status_engine_call();
+        }
+        else {
+            ap_log_error(APLOG_MARK, APLOG_NOTICE, 0, NULL,
+                    "Status engine is currently disabled, enable it by set " \
+                    "SecStatusEngine to On.");
+        }
+#endif
     }
 
     srand((unsigned int)(time(NULL) * getpid()));
@@ -764,7 +801,7 @@ static int hook_request_early(request_rec *r) {
         if (msr->request_content_length > msr->txcfg->reqbody_limit) {
             msr_log(msr, 1, "Request body (Content-Length) is larger than the "
                     "configured limit (%ld).", msr->txcfg->reqbody_limit);
-            if(msr->txcfg->is_enabled != MODSEC_DETECTION_ONLY)
+            if(msr->txcfg->is_enabled != MODSEC_DETECTION_ONLY && msr->txcfg->if_limit_action != REQUEST_BODY_LIMIT_ACTION_PARTIAL)
                 return HTTP_REQUEST_ENTITY_TOO_LARGE;
         }
     }
@@ -1023,7 +1060,7 @@ static void hook_error_log(const char *file, int line, int level, apr_status_t s
 #else
         msr = create_tx_context((request_rec *)r);
 #endif
-        if (msr->txcfg->debuglog_level >= 9) {
+        if (msr != NULL && msr->txcfg->debuglog_level >= 9) {
             if (msr == NULL) {
                 msr_log(msr, 9, "Failed to create context after request failure.");
             }
@@ -1332,29 +1369,30 @@ static int hook_connection_early(conn_rec *conn)
 {
     sb_handle *sb = conn->sbh;
     int i, j;
-    unsigned long int ip_count = 0, ip_count_w = 0;
+    unsigned long int ip_count_r = 0, ip_count_w = 0;
+    char *error_msg;
     worker_score *ws_record = NULL;
 #if AP_SERVER_MAJORVERSION_NUMBER > 1 && AP_SERVER_MINORVERSION_NUMBER > 2
     ap_sb_handle_t *sbh = NULL;
+    char *client_ip = conn->client_ip;
+#else
+    char *client_ip = conn->remote_ip;
 #endif
 
-    if(sb != NULL && (conn_read_state_limit > 0 || conn_write_state_limit > 0))   {
+    if (sb != NULL && (conn_read_state_limit > 0 || conn_write_state_limit > 0)) {
 
         ws_record = &ap_scoreboard_image->servers[sb->child_num][sb->thread_num];
-        if(ws_record == NULL)
+        if (ws_record == NULL)
             return DECLINED;
 
-#if AP_SERVER_MAJORVERSION_NUMBER > 1 && AP_SERVER_MINORVERSION_NUMBER > 2
-        apr_cpystrn(ws_record->client, conn->client_ip, sizeof(ws_record->client));
-#else
-        apr_cpystrn(ws_record->client, conn->remote_ip, sizeof(ws_record->client));
-#endif
+        apr_cpystrn(ws_record->client, client_ip, sizeof(ws_record->client));
+
         for (i = 0; i < server_limit; ++i) {
             for (j = 0; j < thread_limit; ++j) {
 
 #if AP_SERVER_MAJORVERSION_NUMBER > 1 && AP_SERVER_MINORVERSION_NUMBER > 2
                 sbh = conn->sbh;
-                if (sbh == NULL)        {
+                if (sbh == NULL) {
                     return DECLINED;
                 }
 
@@ -1363,27 +1401,19 @@ static int hook_connection_early(conn_rec *conn)
                 ws_record = ap_get_scoreboard_worker(i, j);
 #endif
 
-                if(ws_record == NULL)
+                if (ws_record == NULL)
                     return DECLINED;
 
                 switch (ws_record->status) {
                     case SERVER_BUSY_READ:
-#if AP_SERVER_MAJORVERSION_NUMBER > 1 && AP_SERVER_MINORVERSION_NUMBER > 2
-                        if (strcmp(conn->client_ip, ws_record->client) == 0)
-                            ip_count++;
-#else
-                        if (strcmp(conn->remote_ip, ws_record->client) == 0)
-                            ip_count++;
-#endif
+                        if (strcmp(client_ip, ws_record->client) == 0)
+                            ip_count_r++;
                         break;
+
                     case SERVER_BUSY_WRITE:
-#if AP_SERVER_MAJORVERSION_NUMBER > 1 && AP_SERVER_MINORVERSION_NUMBER > 2
-                        if (strcmp(conn->client_ip, ws_record->client) == 0)
+                        if (strcmp(client_ip, ws_record->client) == 0)
                             ip_count_w++;
-#else
-                        if (strcmp(conn->remote_ip, ws_record->client) == 0)
-                            ip_count_w++;
-#endif
+
                         break;
                     default:
                         break;
@@ -1391,22 +1421,76 @@ static int hook_connection_early(conn_rec *conn)
             }
         }
 
-        if ((conn_read_state_limit > 0) && (ip_count > conn_read_state_limit)) {
-#if AP_SERVER_MAJORVERSION_NUMBER > 1 && AP_SERVER_MINORVERSION_NUMBER > 2
-            ap_log_error(APLOG_MARK, APLOG_WARNING, 0, NULL, "ModSecurity: Access denied with code 400. Too many threads [%ld] of %ld allowed in READ state from %s - Possible DoS Consumption Attack [Rejected]", ip_count,conn_read_state_limit,conn->client_ip);
-#else
-            ap_log_error(APLOG_MARK, APLOG_WARNING, 0, NULL, "ModSecurity: Access denied with code 400. Too many threads [%ld] of %ld allowed in READ state from %s - Possible DoS Consumption Attack [Rejected]", ip_count,conn_read_state_limit,conn->remote_ip);
-#endif
-            return OK;
-        } else if ((conn_write_state_limit > 0) && (ip_count_w > conn_write_state_limit)) {
-#if AP_SERVER_MAJORVERSION_NUMBER > 1 && AP_SERVER_MINORVERSION_NUMBER > 2
-            ap_log_error(APLOG_MARK, APLOG_WARNING, 0, NULL, "ModSecurity: Access denied with code 400. Too many threads [%ld] of %ld allowed in WRITE state from %s - Possible DoS Consumption Attack [Rejected]", ip_count_w,conn_write_state_limit,conn->client_ip);
-#else
-            ap_log_error(APLOG_MARK, APLOG_WARNING, 0, NULL, "ModSecurity: Access denied with code 400. Too many threads [%ld] of %ld allowed in WRITE state from %s - Possible DoS Consumption Attack [Rejected]", ip_count_w,conn_write_state_limit,conn->remote_ip);
-#endif
-            return OK;
-        } else {
-            return DECLINED;
+        if (conn_read_state_limit > 0 && ip_count_r > conn_read_state_limit)
+        {
+            if (conn_read_state_suspicious_list &&
+                (tree_contains_ip(conn->pool,
+                   conn_read_state_suspicious_list, client_ip, NULL, &error_msg) <= 0))
+            {
+                if (conn_limits_filter_state == MODSEC_DETECTION_ONLY)
+                    ap_log_error(APLOG_MARK, APLOG_WARNING, 0, NULL, 
+                        "ModSecurity: Too many threads [%ld] of %ld allowed " \
+                        "in READ state from %s - There is a suspission list " \
+                        "but that IP is not part of it, access granted",
+                        ip_count_r, conn_read_state_limit, client_ip);
+            }
+            else if (tree_contains_ip(conn->pool,
+                conn_read_state_whitelist, client_ip, NULL, &error_msg) > 0)
+            {
+                if (conn_limits_filter_state == MODSEC_DETECTION_ONLY)
+                    ap_log_error(APLOG_MARK, APLOG_WARNING, 0, NULL,
+                        "ModSecurity: Too many threads [%ld] of %ld allowed " \
+                        "in READ state from %s - Ip is on whitelist, access " \
+                        "granted", ip_count_r, conn_read_state_limit,
+                        client_ip);
+            }
+            else
+            {
+                ap_log_error(APLOG_MARK, APLOG_WARNING, 0, NULL,
+                    "ModSecurity: Access denied with code 400. Too many " \
+                    "threads [%ld] of %ld allowed in READ state from %s - " \
+                    "Possible DoS Consumption Attack [Rejected]", ip_count_r,
+                    conn_read_state_limit, client_ip);
+
+                if (conn_limits_filter_state == MODSEC_ENABLED)
+                    return OK;
+            }
+        }
+
+        if (conn_write_state_limit > 0 && ip_count_w > conn_write_state_limit)
+        {
+            if (conn_write_state_suspicious_list &&
+                (tree_contains_ip(conn->pool,
+                    conn_write_state_suspicious_list, client_ip, NULL, &error_msg) <= 0))
+            {
+                if (conn_limits_filter_state == MODSEC_DETECTION_ONLY)
+                    ap_log_error(APLOG_MARK, APLOG_WARNING, 0, NULL,
+                        "ModSecurity: Too many threads [%ld] of %ld allowed " \
+                        "in WRITE state from %s - There is a suspission list " \
+                        "but that IP is not part of it, access granted",
+                        ip_count_w, conn_read_state_limit, client_ip);
+            }
+            else if (tree_contains_ip(conn->pool,
+                conn_write_state_whitelist, client_ip, NULL, &error_msg) > 0)
+            {
+                if (conn_limits_filter_state == MODSEC_DETECTION_ONLY)
+                    ap_log_error(APLOG_MARK, APLOG_WARNING, 0, NULL,
+                        "ModSecurity: Too many threads [%ld] of %ld allowed " \
+                        "in WRITE state from %s - Ip is on whitelist, " \
+                        "access granted", ip_count_w, conn_read_state_limit,
+                        client_ip);
+            }
+            else
+            {
+                ap_log_error(APLOG_MARK, APLOG_WARNING, 0, NULL,
+                    "ModSecurity: Access denied with code 400. Too many " \
+                    "threads [%ld] of %ld allowed in WRITE state from %s - " \
+                    "Possible DoS Consumption Attack [Rejected]", ip_count_w,
+                    conn_write_state_limit, client_ip);
+
+                if (!conn_limits_filter_state == MODSEC_ENABLED)
+                    return OK;
+            }
         }
     }
 
@@ -1468,6 +1552,7 @@ static void register_hooks(apr_pool_t *mp) {
     static const char *const postread_beforeme_list[] = {
         "mod_rpaf.c",
         "mod_rpaf-2.0.c",
+        "mod_extract_forwarded.c",
         "mod_extract_forwarded2.c",
         "mod_remoteip.c",
         "mod_custom_header.c",
